import os
import sys
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")
from timeit import default_timer
import numpy as np, matplotlib.pyplot as pp

for module in [('euphonic', 'data', 'interpolation'), ('brille', 'euphonic')]:
    for i in range(len(module)):
        check = '.'.join(np.array(module)[0:i+1])
        if find_spec(check) is None:
            raise Exception('Required module {} not found'.format(check))
from euphonic.data.interpolation import InterpolationData
from brille.euphonic import BrEu
import brille

def load_interpolation_data(named):
    """Load a data file from the repository tests folder

    Note that this only works if the package is installed via
        python setup.py develop [--user]
    which builds the C++ library and installs it inside of the repository
    structure with simlinks in the usual package location(s).
    """
    test_spec = find_spec('brille')
    brilleroot = test_spec.submodule_search_locations[0]
    seed = os.path.join(brilleroot,'..','tests',named)
    return InterpolationData(seed)

class timer:
    def __init__(self):
        self.tic()
    def tic(self):
        self.times_called = 0
        self.start_time = default_timer()
        self.elapsed_time = 0
    def toc(self):
        self.times_called += 1
        self.elapsed_time =  default_timer()-self.start_time
        return self.elapsed_time
    def elapsed(self):
        return self.elapsed_time
    def average(self):
        return self.elapsed_time/self.times_called
    def uncertainty(self):
        return np.sqrt(self.average())/self.times_called
    def relative_uncertainty(self):
        if self.times_called < 1:
            return np.inf
        return 1/np.sqrt(self.times_called*self.elapsed_time)
tictoc = timer()

def cen2corner(Z):
    dX = np.diff(Z, axis=0)/2
    Z = np.concatenate((
        Z[:-1]-dX,
        Z[np.newaxis, -1]-dX[np.newaxis, -1],
        Z[np.newaxis, -1]+dX[np.newaxis, -1]), axis=0)
    dY = np.diff(Z, axis=1)/2
    Z = np.concatenate((
        Z[:, :-1]-dY,
        Z[:, np.newaxis, -1]-dY[:, np.newaxis, -1],
        Z[:, np.newaxis, -1]+dY[:, np.newaxis, -1]), axis=1)
    return Z

def pcolormesh(X,Y,Z, *args,**kwargs):
    return pp.pcolormesh(cen2corner(X), cen2corner(Y), Z, *args, **kwargs)

nb = load_interpolation_data('nb')

q = np.random.rand(10000,3)*10

senb = BrEu(nb, mesh=True, parallel=True, max_size=0.0001, num_levels=5)

tictoc.tic()
while not (tictoc.elapsed() > 60 or tictoc.relative_uncertainty() < 0.05):
    senb.w_q(q, interpolate=False)
    tictoc.toc()
eu_time_point = tictoc.average()/q.shape[0]*10**6
eu_delt_point = tictoc.uncertainty()/q.shape[0]*10**6

nthreads = np.arange(1,os.cpu_count()//2+1,dtype='int')
it = np.nditer([nthreads, None, None], op_dtypes=('int','double','double'))
for n, t, u in it:
    tictoc.tic()
    while not (tictoc.elapsed() > 20 or tictoc.relative_uncertainty() < 0.05):
        senb.w_q(q, threads=n)
        tictoc.toc()
    print('{} threads: {:6.3f}/{} -> {:5.3f} +/- {:5.3f}'.format(n, tictoc.elapsed(), tictoc.times_called, tictoc.average(), tictoc.uncertainty()))
    t[...]=tictoc.average()
    u[...]=tictoc.uncertainty()
time_point = it.operands[1]/q.shape[0]*10**6
uncertainty_point = it.operands[2]/q.shape[0]*10**6

pp.figure()
pp.errorbar([1],eu_time_point, yerr=eu_delt_point, marker='s', label='Euphonic')
pp.errorbar(nthreads, time_point, yerr=uncertainty_point, marker='o', color='k', linestyle='')
pp.xlabel('Number of threads')
pp.ylabel('t/μs')
pp.title(r'Time to calculate $\epsilon$(Q) and $\omega$(Q) per Q point')
pp.legend()


ticks = [1/x for x in range(1,7)]
labs = [r'$'+'{}'.format(x)+r'^{-1}$' for x in range(1,7)]
pp.figure()
pp.errorbar(1/nthreads, time_point/max(time_point), yerr=uncertainty_point/max(time_point), marker='o', color='k', linestyle='')
pp.setp(pp.gca(),'xticks',ticks,'xticklabels',labs,'yticks',ticks,'yticklabels',labs)
pp.xlabel('inverse number of threads')
pp.title('Relative interpolation time per point')

pp.show()



"""Interface to VisPy for OpenGL visualisation

See `VisPy <https://vispy.org>`_ for installation and configuration directions
"""

from dataclasses import dataclass
from brille import Polyhedron
from vispy.color import Color
from typing import List


@dataclass
class VisPolyhedron:
    """Wrapper to contain visualisation options for a brille.Polyhedron or brille.LPolyhedron

    Attributes
    ----------
    polyhedron : Union[Polyhedron, LPolyhedron]
        The polyhedron to be visualised. A LPolyhedron will be converted automatically to its Cartesian equivalent.
    face_color : Union[str, vispy.color.Color]
        The face color of the visualised polyhedron, default 'black'
    edge_color : Union[str, vispy.color.Color]
        The edge color of the visualised polyhedron, default 'black'
    fill : bool
        Control drawing of the faces, default True
    outline : bool
        Control drawing of the edges, default True

    Note
    ----
    In the case that a polyhedron with vertices expressed in units of a lattice basis is provided, the oriented lattice
    basis vectors will be used to convert the vertices into an orthonormal frame during object initialisation.
    """

    polyhedron: Polyhedron
    face_color: Color = Color("black")
    edge_color: Color = Color("black")
    fill: bool = True
    outline: bool = True
    opacity: float = 0.2

    def __post_init__(self):
        from brille import LPolyhedron

        if isinstance(self.polyhedron, LPolyhedron):
            self.polyhedron = self.polyhedron.to_Cartesian()
        if isinstance(self.face_color, str):
            self.face_color = Color(self.face_color)
        if isinstance(self.edge_color, str):
            self.edge_color = Color(self.edge_color)

    def box(self):
        """Return the minimum and maximum corners of the bounding box"""
        return self.polyhedron.vertices.min(axis=0), self.polyhedron.vertices.max(
            axis=0
        )


def vis_polyhedron(polyhedron: VisPolyhedron, **kwargs):
    """Visualise a single wrapped polyhedron

    Parameters
    ----------
    polyhedron : Union[VisPolyhedron, brille.Polyhedron, brille.LPolyhedron]
        The polyhedron to be visualised. Any object with similar `vertices` and `faces` properties may work as well.
    kwargs
        Optional keyword arguments for the VisPolyhedron constructor. Only used if a Polyhedron or LPolyhedron
        is provided as input.
        face_color : Union[str, vispy.color.Color]
            The face color used for Polyhedron or LPolyhedron object input.
        edge_color : Union[str, vispy.color.Color]
            The edge color used for Polyhedron or LPolyhedron object input.
        fill : bool
            Control drawing the faces of the Polyhedron or LPolyhedron object.
        outline : bool
            Control drawing the edges of the Polyhedron or LPolyhedron object.
    """
    if not isinstance(polyhedron, VisPolyhedron):
        polyhedron = VisPolyhedron(polyhedron, **kwargs)
    vis_polyhedra([polyhedron])


def vis_polyhedra(polyhedra: List[VisPolyhedron], **kwargs):
    """Visualise multiple polyhedra

    Parameters
    ----------
    polyhedra : List[Union[VisPolyhedron, brille.Polyhedron, brille.LPolyhedron]]
        The polyhedron to be visualised. Any object with similar `vertices` and `faces` properties may work as well.
    kwargs
        Optional keyword arguments for the VisPolyhedron constructor. Only used if a Polyhedron or LPolyhedron
        is provided in the polyhedron list.
        face_color : Union[List[Union[str, vispy.color.Color]],Union[str, vispy.color.Color]]
            A single face color used for all Polyhedron and LPolyhedron objects or a list of colors which will be tiled
            to the size of the full polyhedra list.
        edge_color : Union[List[Union[str, vispy.color.Color]],Union[str, vispy.color.Color]]
            A single edge color used for all Polyhedron and LPolyhedron objects or a list of colors which will be tiled
            to the size of the full polyhedra list.
        fill : Union[List[bool], bool]
            A single value to control drawing the faces of all Polyhedron and LPolyhedron objects or a list of boolean
            values which will be tiled to the size of the full polyhedra list.
        outline : Union[List[bool], bool]
            A single value to control drawing the edges of all Polyhedron and LPolyhedron objects or a list of boolean
            values which will be tiled to the size of the full polyhedra list.
        opacity : Union[List[float], float]
            A single value to control face opacity of all Polyhedron and LPolyhedron objects or a list of floating point
            values which will be tiled to the size of the full polyhedra list.
    """
    if any([not isinstance(x, VisPolyhedron) for x in polyhedra]):
        n_poly = len(polyhedra)
        face_colors = make_colours(n_poly, color=kwargs.get("face_color", None))
        edge_colors = make_colours(n_poly, color=kwargs.get("edge_color", None))
        fill = make_list(n_poly, kwargs.get("fill", True))
        outline = make_list(n_poly, kwargs.get("outline", True))
        # opacity = make_list(n_poly, kwargs.get('opacity', 0.2))
        polyhedra = [
            p
            if isinstance(p, VisPolyhedron)
            else VisPolyhedron(p, face_color=fc, edge_color=ec, fill=f, outline=o)
            for p, fc, ec, f, o in zip(
                polyhedra, face_colors, edge_colors, fill, outline
            )
        ]
    from vispy import scene, app

    bg = Color("white")
    canvas = scene.SceneCanvas(keys="interactive", bgcolor=bg, size=(800, 600))
    view = canvas.central_widget.add_view()
    view.camera = "arcball"

    for p in polyhedra:
        if p.fill:
            view.add(
                vis_polyhedron_to_mesh(
                    p.polyhedron, color=p.face_color, opacity=p.opacity
                )
            )
        if p.outline:
            [
                view.add(line)
                for line in vis_polyhedron_boundary(p.polyhedron, color=p.edge_color)
            ]

    # set the camera range based on properties of the polyhedra...
    from numpy import vstack

    boxes = [p.box() for p in polyhedra]
    box_min = vstack([m for m, _ in boxes]).min(axis=0)
    box_max = vstack([m for _, m in boxes]).max(axis=0)
    view.camera.set_range(
        x=[box_min[0], box_max[0]],
        y=[box_min[1], box_max[1]],
        z=[box_min[2], box_max[2]],
    )

    # finalise the VisPy setup and execute
    canvas.show()
    app.run()


def vis_polyhedron_to_mesh(polyhedron: Polyhedron, color=None, opacity=1.0):
    """Construct as VisPy mesh from a Polyhedron object"""
    from vispy.scene.visuals import Mesh
    from numpy import array

    if color is None:
        color = "red"
    v = polyhedron.vertices.astype("float32")
    faces = polyhedron.faces
    # ensure the faces are all triangulated:
    faces = [
        [face[0], face[i], face[(i + 1) % len(face)]]
        for face in faces
        for i in range(1, len(face) - 1)
    ]
    mesh = Mesh(v, array(faces, dtype="uint32"), shading="flat", color=color)
    mesh.opacity = opacity
    return mesh


def vis_polyhedron_boundary(polyhedron: Polyhedron, color=None):
    """Construct a list of VisPy Line objects from the edges of a Polyhedron object"""
    from vispy.scene import Line

    v, faces = polyhedron.vertices.astype("float32"), polyhedron.faces
    if color is None:
        color = "black"
    lines = [
        Line(v[[*face, face[0]]], color=color, method="gl", width=3, antialias=True)
        for face in faces
    ]
    return lines


def make_colours(n, color=None):
    """Construct a list of colors for use in displaying Polyhedron objects

    Parameters
    ----------
    n : int
        The number of colors required
    color : Union[List[Union[str, vispy.color.Color]], Union[str, vispy.color.Color]]
        If color is not provided, the list of all colors known to VisPy will be used

    Returns
    -------
    List[vispy.color.Color]
        A length-n list of colors. If the starting value of `color` is less than length-n, it will be tiled to length-n.

    Examples
    --------
    >>> make_colours(7, ['red', 'blue', 'green'])
    [Color('red'), Color('blue'), Color('green'), Color('red'), Color('blue'), Color('green'), Color('red')]

    >>> from vispy.color import Color
    >>> make_colors(4, Color('black'))
    [Color('black'), Color('black'), Color('black'), Color('black')]
    """
    if color is None:
        from vispy.color import get_color_names

        color = filter(lambda x: len(x) > 1, get_color_names())

    from collections.abc import Iterable

    if isinstance(color, Iterable):
        color = list(color)
    if isinstance(color, str) or (isinstance(color, (list, tuple)) and len(color) == 3):
        color = [color]

    from numpy import ndarray, array, tile

    if not isinstance(color, ndarray):
        color = array(color)
    if color.shape[0] < n:
        color = tile(color, 1 + n // color.shape[0])
    return color[0:n]


def make_list(n, val):
    """Construct a 1-D list of values for use in displaying Polyhedron objects

    Parameters
    ----------
    n : int
        The number of colors required
    val : Union[List[T], T]

    Returns
    -------
    List[T]
        A length-n list of values. If the starting value is less than length-n, it will be tiled to length-n.
        If the input list is longer than the requested list length, it will be truncated.

    Examples
    --------
    >>> make_list(10, [True, False, True, True])
    [True, False, True, True, True, False, True, True, True, False]

    The items in the input list should all be the same type. In the case of numeric types, the numpy type promotion
    rules will cast them all to a consistent type before tiling or truncating. So an 'unused' input value may influence
    the type of the returned values.

    >>> make_list(2, [2, 101, 3.14159])
    [2.0, 101.0]
    """
    from collections.abc import Iterable

    if isinstance(val, Iterable):
        val = list(val)
    from numpy import ndarray, array, tile

    if not isinstance(val, ndarray):
        val = array(val)
    if val.size < n:
        val = tile(val, 1 + n // val.size)
    return val[:n]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))
import brille._brille as brille_module

# -- Project information -----------------------------------------------------

project = "brille"
copyright = "2020, Gregory Tucker"
author = "Gregory Tucker"

version = brille_module.__version__  # just the 'short' version
release = brille_module.version  # the 'full' version information


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "breathe",
    "sphinxcontrib.katex",
    "sphinxcontrib.tikz",
    #'exhale',
]

# Useful mappings: https://gist.github.com/bskinn/0e164963428d4b51017cebdb6cda5209
intersphinx_mapping = {
    "euphonic": ("https://euphonic.readthedocs.io/en/stable/", None),
    "brilleu": ("https://brille.github.io/brilleu/latest/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# The image now has sphinxcontrib-katex v0.9.10, which uses katex 0.16.10
# # Some :math:`[LaTeX]` directives insert '\r' into the string passed to katex?
# # This raises an error with the version specified in the Docker image's
# # sphinxcontrib.katex, 0.11.1, switching to v0.15.0 turns the error into
# # a warning about 'LaTeX-incompatible input'
# # Updating the Docker image might be possible, but the exhale/breathe/sphinx
# # problem encountered outside of the Docker build would need to be avoided.
# katex_css_path = 'https://cdn.jsdelivr.net/npm/katex@0.15.0/dist/katex.min.css'

napoleon_use_ivar = True
napoleon_use_param = False
napoleon_use_admonition_for_notes = True

tikz_proc_suite = "pdf2svg"  # We are building exclusively in our own container, no ReadTheDocs ghostscript restriction

breathe_projects = {"brille": "_build/doxygenxml/"}
breathe_default_project = "brille"
# The naming of this breath directive seems backwards, but for it to make any sense
# the keys must be the extensions and their values the associated domain(s)
breathe_domain_by_extension = {"h": "cpp", "hpp": "cpp", "tpp": "cpp"}

# # setup the exhale extension
# exhale_args = {
#     # required arguments first:
#     "containmentFolder": "./api",
#     "rootFileName": "library_root.rst",
#     "rootFileTitle": "C++ Library API",
#     "doxygenStripFromPath": "..",
#     # Suggested optional arguments
#     "createTreeView": True,
#     # TIP: if using the sphinx-bootstrap-theme, you also need
#     # "treeViewIsBootstrap": True,
#     "exhaleExecutesDoxygen": True,
#     #"exhaleSilentDoxygen": True,
#     "exhaleUseDoxyfile": True,
#     # "exhaleDoxygenStdin": "INPUT = ../wrap/", # "INPUT = ../src/",
# }

autosummary_generate = True

autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

source_suffice = ".rst"
master_doc = "index"


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
# exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
exclude_patterns = ["Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# # following github.com/pybind/pybind11/blob/stable/docs/conf.py:
# this_is_rtd = os.environ.get('READTHEDOCS',None) == 'True'
# if this_is_rtd:
#     html_context = {'css_files' :[
#         '//media.readthedocs.org/css/sphinx_rtd_theme.css',
#         '//media.readthedocs.org/css/readthedocs-doc-embed.css',
#         '_static/theme_overrides.css'
#     ]}
# else:
#     import sphinx_rtd_theme
#     html_theme = 'sphinx_rtd_theme'
#     html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
#     html_context = {'css_files': ['_static/theme_overrides.css']}

# Alternatively, just specify a theme:
html_theme = "sphinx_rtd_theme"
html_logo = "../brille.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def call_and_check(command, **kwds):
    from subprocess import call, CalledProcessError

    try:
        retcode = call(command, **kwds)
        if retcode < 0:
            sys.stderr.write(f"{call} error code: {-retcode}\n")
    except (CalledProcessError, OSError) as e:
        sys.stderr.write(f"{call} execution failed: {e}\n")


# again, following from github.com/pybind/pybind11/blob/stable/docs/conf.py:
def generate_doxygen_xml(app):
    build_dir = os.path.join(app.confdir, "_build")
    if not os.path.exists(build_dir):
        os.mkdir(build_dir)

    call_and_check("doxygen", cwd=app.confdir)
    call_and_check(
        [
            "breathe-apidoc",
            "--output-dir=_build/breathe",
            "-f",
            "-g",
            "class,namespace",
            "_build/doxygenxml/",
        ],
        cwd=app.confdir,
    )


def setup(app):
    """Add hook for building Doxygen xml when needed"""
    app.connect("builder-inited", generate_doxygen_xml)



# Copyright © 2020 Greg Tucker <greg.tucker@stfc.ac.uk>
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""Python module :py:mod:`brille`
=================================

This module provides access to the C++ brille library which can be used to
interact with spacegroup and pointgroup symmetry operations, to determine the
first Brillouin zone for a given real space crystallographic lattice, to
find *an* irreducible polyhedron from the first Brillouin zone and the
pointgroup operations of a spacegroup, to construct polyhedron-filling connected
point networks, and to perform linear-interpolation of user-provided data for
any reciprocal space point at a symmetry-equivalent position within the
connected point network.

.. currentmodule:: brille

.. autosummary::
    :toctree: _generate
"""

from .bound import (
    __version__,
    version,
    AngleUnit,
    LengthUnit,
    Bravais,
    RotatesLike,
    NodeType,
    real_space_tolerance,
    reciprocal_space_tolerance,
    ApproxConfig,
    Basis,
    HallSymbol,
    Spacegroup,
    Pointgroup,
    Symmetry,
    PointSymmetry,
    PrimitiveTransform,
    BrillouinZone,
    Polyhedron,
    LPolyhedron,
    SortingStatus,
    BZTrellisQcc,
    BZTrellisQdc,
    BZTrellisQdd,
    BZMeshQcc,
    BZMeshQdc,
    BZMeshQdd,
    BZNestQcc,
    BZNestQdc,
    BZNestQdd,
    __grid_types__,
)
from .lattice import Lattice
from . import utils
from . import plotting


__all__ = [
    "__version__",
    "version",
    "AngleUnit",
    "LengthUnit",
    "Bravais",
    "RotatesLike",
    "NodeType",
    "real_space_tolerance",
    "reciprocal_space_tolerance",
    "ApproxConfig",
    "Basis",
    "HallSymbol",
    "Spacegroup",
    "Pointgroup",
    "Symmetry",
    "PointSymmetry",
    "PrimitiveTransform",
    "BrillouinZone",
    "Polyhedron",
    "LPolyhedron",
    "SortingStatus",
    "BZTrellisQcc",
    "BZTrellisQdc",
    "BZTrellisQdd",
    "BZMeshQcc",
    "BZMeshQdc",
    "BZMeshQdd",
    "BZNestQcc",
    "BZNestQdc",
    "BZNestQdd",
    "__grid_types__",
    "Lattice",
    "utils",
    "plotting",
]



def _t(args):
    return args if isinstance(args, tuple) else tuple(args)


def _make(to_type, args):
    return args if isinstance(args, to_type) else to_type(*_t(args))


# def Lattice(values, /, *, spacegroup=None, symmetry=None, basis=None, **kwargs):
def Lattice(values, spacegroup=None, symmetry=None, basis=None, **kwargs):
    """Construct a space-spanning lattice in three dimensions

    A space-spanning lattice in :math:`N` dimensions has :math:`N` basis vectors
    which can be described fully by their :math:`N` lengths and the
    :math:`\\frac{1}{2} N (N-1)` angles between each set of basis vectors, or
    :math:`\\frac{1}{2}N(N+1)` scalars in total.
    This class stores the basis vectors of the lattice described in an orthonormal space,
    plus the metric of the space, and the equivalent information for the dual of the lattice.

    Examples
    --------

    >>> from brille import Lattice
    >>> a, c = 3.95, 12.9
    >>> avec, bvec, cvec = [a, 0, 0], [0, a, 0], [0, 0, c]
    >>> basis = ([[0, 0, 0], [0.5, 0.5, 0.5], [0.25, 0.25, 0.3]], [0, 0, 1]))
    >>> symmetry = (rotations, translations)
    >>> lat = Lattice(([a, a, c], [90, 90, 90]), spacegroup='I4/mmm', basis=basis)
    >>> lat = Lattice(([avec, bvec, cvec],), symmetry=symmetry, basis=basis)
    >>> lat = Lattice([avec, bvec, cvec], spacegroup='I4/mmm', basis=basis)

    Note
    ----
    The parameters packed into ``values`` are position-based and should be *either* ``(lengths, angles)``
    *or* ``(vectors,)`` [which also requires setting ``row_vectors=False`` if the matrix represents column vectors].
    The lengths or vector components are interpreted in angstrom or inverse angstrom for real or reciprocal lattice
    parameters, respectively, and are assumed to be angstrom if the keyword ``real_space`` is missing or True.


    Parameters
    ----------

    values: (lengths, angles), vectors
        See the note above and the documenation for ``lengths``, ``angles``, and ``vectors`` below.
    lengths : list, tuple, numpy.ndarray
        The three basis vector lengths in angstrom for real lattice, or inverse angstrom for reciprocal lattices
    angles : list, tuple, numpy.ndarray
        The three angles between the basis vectors in degrees or radians. The angles are interpreted as radians
        if none are greater than pi and are otherwise assumed to be degrees.
    vectors: list[list,...], tuple(tuple,...), numpy.ndarray
        The basis vectors in angstrom for real lattices, or inverse angstrom for reciprocal lattices,
        expressed in an orthorhombic coordinate system. An optional keyword argument, ``row_vectors``, identifies
        if the provided basis vectors are row vectors [`row_vectors=True`, default] or column vectors
        [``row_vectors=False``].
    spacegroup : str, tuple(str, str)
        The International Tables name, Hermann-Mauguin symbol with optional choice, or Hall symbol for the
        spacegroup of the lattice. The spacegroup may be provided as positional argument(s) or by keyword.
        If present, the ``symmetry`` keyword must not be used.
        Valid syntax for ``(Hermann-Mauguin, choice)`` input depends on the spacegroup but is generally one of:

        * a single letter ('a', 'b', or 'c') with possible prepended '-', denoting unique-axis choice
        * a single digit ('1', '2', or '3'), denoting origin choice
        * a letter and digit, denoting unique-axis and origin choice
        * a permutation of 'abc' with possible '-' before one of the letters, denoting axis permutation
        * or 'R' or 'H' for trigonal systems with Rhombohedral or Hexagonal lattice settings, respectively.

        Acceptable values are contained in the C++ source code in the seventh column of
        `this table <https://github.com/brille/brille/blob/eecb4cb28227665908793abc47e88c69518c09fc/src/spg_database.cpp#L63-L610>`_
        with each line representing one spacegroup with values, in order, defined by the
        `class signature <https://github.com/brille/brille/blob/eecb4cb28227665908793abc47e88c69518c09fc/src/spg_database.hpp#L80-L91>`_
        These values come from spglib, which likely obtained them from
        `Seto's Home Page <https://web.archive.org/web/20210621195003/http://pmsl.planet.sci.kobe-u.ac.jp/~seto/?page_id=37&lang=en>`_.
    symmetry: brille.Symmetry, tuple(matrices, vectors), str
        The spacegroup symmetry operations as an object, a tuple of the (pseudo)rotation matrices and translation
        vectors, or a CIF xyz encoded string. The symmetry information must be provided by keyword.
        If present, the 'spacegroup' positional argument(s) or keyword must not be used.
    basis: brille.Basis, tuple(positions, types)
        The atom basis information of the lattice, expressed in units of the real space basis vectors. The types
        must be integer and are used only to identify equivalent atoms -- they should probably be contiguous from
        zero to 1-N where N is the number of unique atoms in the atom basis.
        If present, either spacegroup or symmetry information must be provided.
    kwargs:
        Keyword arguments are passed to the :py:class:`brille._brille.Lattice` constructor,
        see its documentation for details.

    """
    from ._brille import Lattice as _Lattice, Symmetry as _Symmetry, Basis as _Basis

    if isinstance(spacegroup, str):
        spacegroup = (spacegroup,)
    if spacegroup is None:
        spacegroup = tuple()
    if len(spacegroup) and not all([isinstance(x, str) for x in spacegroup]):
        raise ValueError(
            "Keyword argument spacegroup should be one of: str, (str,), or (str, str)"
        )

    if basis is not None and symmetry is None and len(spacegroup) == 0:
        raise ValueError(
            "Providing basis without symmetry or spacegroup is not allowed"
        )

    if symmetry is not None and len(spacegroup):
        raise ValueError("Providing both spacegroup and symmetry is not allowed")

    # The pybind11 wrapper has overloads which accept:
    #   3-tuple-like, 3-tuple-like : 3-element arrays of lengths and angles
    #   3-tuple(3-tuple-like)-like : 3 3-element arrays of basis vectors
    #
    # with the basis vectors ((ax, ay, az), (bx, by, bz), (cx, cy, cz)) or ((ax, bx, cx), (ay, by, cy), (az, bz, cz))
    #
    # If the user provided a 2-tuple then it must be (lengths, angles), otherwise it must be vectors
    # We use less-than to allow a user to provide (vectors,) as the first input
    args = values if len(values) < 3 else (values,)

    if symmetry is None and basis is None:
        return _Lattice(*args, *spacegroup, **kwargs)
    elif symmetry is None:
        return _Lattice(*args, *spacegroup, _make(_Basis, basis), **kwargs)

    if basis is None:
        return _Lattice(*args, _make(_Symmetry, symmetry), **kwargs)
    else:
        return _Lattice(
            *args, _make(_Symmetry, symmetry), _make(_Basis, basis), **kwargs
        )



# Copyright © 2019,2020 Greg Tucker <greg.tucker@stfc.ac.uk>
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""
Plotting utilities for ``brille``
---------------------------------

.. currentmodule:: brille.plotting

.. autosummary::
    :toctree: _generate
"""

from importlib.util import find_spec

__all__ = []

if find_spec("matplotlib") is None:
    print("Matplotlib not found; plotting functions will not be available")
else:
    import numpy as np
    import matplotlib.pyplot as pp
    from mpl_toolkits.mplot3d.axes3d import Axes3D
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection, Line3DCollection

    __all__.extend(
        [
            "plot",
            "plot_points",
            "plot_points_with_lines",
            "plot_bz",
            "plot_polyhedron",
            "plot_tetrahedron",
            "plot_tetrahedra",
            "make_colours",
        ]
    )

    def _check_axes(axs=None):
        if axs is None:
            if pp.get_fignums() and isinstance(pp.gca(), Axes3D):
                axs = pp.gca()
            else:
                axs = Axes3D(pp.figure())
            pp.gcf().add_axes(axs)
        return axs

    def plot(*args, **kwds):
        """A gateway plotting function which infers intention from its input.

        There are a number of plotting routines within this module that have names
        associated with their intended input and functionality. Their signatures
        are sufficiently distinct that the intended plotting function can be
        determined by examining the calling signature alone.
        This function takes any input and calls one of :py:func:`plot_bz`,
        :py:func:`plot_polyhedron`, :py:func:`plot_points`,
        :py:func:`plot_points_with_lines`, or :py:func:`plot_tetrahedra`, depending
        on the input provided.

        Parameters
        ----------
        *args
            Variable length argument list, used exclusively for determining the
            implied plotting specialisation.
        **kwargs
            Arbitrary keyword arguments, passed unmodified to the implied
            specialisation.

        Returns
        -------
        variable
            The return type depends on which specialisation is called.

        Raises
        ------
        Exception
            If the specialisation can not be inferred from ``*args`` then an
            exception is raised.
        """
        from .bound import BrillouinZone, Polyhedron, __grid_types__

        if len(args) == 1:
            if isinstance(args[0], (BrillouinZone, *__grid_types__)):
                return plot_bz(*args, **kwds)
            if isinstance(args[0], Polyhedron):
                return plot_polyhedron(*args, **kwds)
            else:
                return plot_points(*args, **kwds)
        if len(args) == 2:
            if (isinstance(args[1], np.ndarray)) and not issubclass(
                args[1].dtype.type, np.integer
            ):
                return plot_points_with_lines(*args, **kwds)
            else:
                return plot_tetrahedra(*args, **kwds)
        else:
            raise Exception("Unknown number of non-keyword arguments for plot")

    def plot_points(x, axs=None, title=None, show=True):
        """Plot points.

        Parameters
        ----------
        x : :py:class:`numpy.ndarray`
            A :math:`N \\times 3` two dimensional array of :math:`N` points to plot.
        axs : :py:class:`matplotlib.axes.Axes`, optional
            The axes in which to add the plotted points. If ``None`` then
            :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
            axes.
        title : str, optional
            An optional title for the plotting axes `axs`
        show : bool, optional
            Whether to call `matplotlib.pyplot.show()` after adding the points to
            `axs`; this is mostly useful in non-interactive environments.
        """
        axs = _check_axes(axs)
        axs.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        if title is not None:
            axs.set_title(title)
        if show:
            pp.show()

    def plot_points_with_lines(x, y, axs=None, title=None, show=True):
        """Plot points with lines.

        Parameters
        ----------
        x : :py:class:`numpy.ndarray`
            A :math:`N \\times 3` two dimension array of :math:`N` points to plot.
        y : :py:class:`numpy.ndarray`
            A :math:`(M+1) \\times 3` two dimensional array of the endpoints of :math:`M` connected
            line segments to plot.
        axs : :py:class:`matplotlib.axes.Axes`, optional
            The axes in which to add the plotted points. If ``None`` then
            :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
            axes.
        title : str, optional
            An optional title for the plotting axes `axs`
        show : bool, optional
            Whether to call `matplotlib.pyplot.show()` after adding the points to
            `axs`; this is mostly useful in non-interactive environments.

        Note
        ----
        The :math:`M` line segments defined by `y` are drawn before the points in `x`.
        """
        axs = _check_axes(axs)
        axs.plot(y[:, 0], y[:, 1], y[:, 2])
        axs.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        if title is not None:
            axs.set_title(title)
        if show:
            pp.show()

    def plot_bz(
        bz,
        axs=None,
        origin=None,
        Q=None,
        units="invA",
        irreducible=True,
        face_vectors=False,
        show=True,
        color="b",
        edgecolor="k",
        linewidth=1,
        alpha=0.2,
    ):
        """Plot a :py:class:`BrillouinZone` or related object.

        Draw the faces of a first Brillouin zone and/or irreducible Brillouin zone
        polyhedron, plus additional structures.

        Parameters
        ----------
        bz : :py:class:`BrillouinZone`, \
             :py:class:`BZMeshQdc`, :py:class:`BZMeshQcc`, :py:class:`BZMeshQdd`, \
             :py:class:`BZNestQdc`, :py:class:`BZNestQcc`, :py:class:`BZNestQdd`, \
             :py:class:`BZTrellisQdc`, :py:class:`BZTrellisQcc`, :py:class:`BZTrellisQ`
            The object containing information about a first Brillouin zone and/or
            an irreducible Brillouin zone.

        axs : :py:class:`matplotlib.axes.Axes`, optional
            The axes in which to add the plotted points. If ``None`` then
            :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
            axes.

        origin : {:py:class:`numpy.ndarray`,tuple,list}, optional
            The origin of the plotting coordinate system, all drawn information is
            relative to this vector. Any 3-element object convertible to a
            :py:class:`numpy.ndarray` is valid input. Invalid input is replaced by
            the zero-vector.

        Q : :py:class:`numpy.ndarray`, optional
            A :math:`N \\times 3` array of points to draw after the first/irreducible
            Brillouin zone polyhedron. If `bz` is not a :py:class:`BrillouinZone`
            and input `Q` is ``None``, `Q` will be replaced by the contents of
            `bz.rlu` or `bz.invA`, depending on the value of `units`; otherwise
            the units of `Q` are assumed to be the same as `units`.

        units : str, optional
            The units in which to plot the first/irreducible Brillouin zone.

            +-----------------+---------------------------------------------------+
            | valid units     | corresponding to                                  |
            +=================+===================================================+
            | ``'invA'``      | inverse ångstrom                                  |
            +-----------------+---------------------------------------------------+
            | ``'rlu'``       | reciprocal lattice units of the conventional cell |
            +-----------------+---------------------------------------------------+
            | ``'primitive'`` | reciprocal lattice units of the primitive cell    |
            +-----------------+---------------------------------------------------+

        irreducible : bool, optional
            Whether to plot the irreducible Brillouin zone polyhedron when it is
            present. When ``True``, the first Brillouin zone edges are plotted
            as well.

        face_vectors : bool, optional
            Whether to plot vectors from the origin through each first Brillouin
            zone face centre.

        show : bool, optional
            Whether to call `matplotlib.pyplot.show()` after adding the points to
            `axs`; this is mostly useful in non-interactive environments.

        color : optional
            The face color of the drawn polyhedra.

        edgecolor : optional
            The edge color of the drawn polyhedra.

        linewidth : float, optional
            The edge line width of drawn polyhedra.

        alpha : float, optional
            The face alpha of drawn polyhedra.

        Returns
        -------
        :py:class:`matplotlib:axes:Axes`
            The value of `axs` after plotting.
        """
        from .bound import __grid_types__

        axs = _check_axes(axs)
        if isinstance(bz, __grid_types__):
            if Q is None:
                if units == "rlu":
                    Q = bz.rlu
                elif units == "invA":
                    Q = bz.invA
            bz = bz.BrillouinZone
        if origin is not None and not isinstance(origin, np.ndarray):
            origin = np.array(origin)
        if origin is None or origin.size != 3 or origin.ndim > 1:
            origin = np.array((0, 0, 0))
        # we always draw the 1st Brillouin zone
        if units == "rlu":
            verts = bz.vertices
        elif units == "primitive":
            verts = bz.vertices_primitive
        else:
            verts = bz.vertices_invA
        bzcolor = color if not irreducible else "w"
        bzedgecolor = edgecolor if not irreducible else "0.5"
        bzlinestyle = "-" if not irreducible else "--"
        bzalpha = alpha if not irreducible else 0

        # the 1st Brillouin zone has on-face points equal to half the normals
        polybz, xyz_min, xyz_max = _make_poly_collection(
            verts,
            bz.vertices_per_face,
            origin=origin,
            color=bzcolor,
            edgecolor=bzedgecolor,
            linestyle=bzlinestyle,
            linewidth=linewidth,
            alpha=bzalpha,
        )
        if irreducible:
            if units == "rlu":
                ir_verts = bz.ir_vertices
            elif units == "primitive":
                ir_verts = bz.ir_vertices_primitive
            else:
                ir_verts = bz.ir_vertices_invA
            if ir_verts.size > 0:
                polyir, _, _ = _make_poly_collection(
                    ir_verts,
                    bz.ir_vertices_per_face,
                    origin=origin,
                    color=color,
                    edgecolor=edgecolor,
                    linestyle="-",
                    linewidth=linewidth,
                    alpha=alpha,
                )
                axs.add_collection3d(polyir)
        axs.add_collection3d(polybz)
        if face_vectors:
            if units == "rlu":
                norms = bz.normals
                point = bz.points
            elif units == "primitive":
                norms = bz.normals_primitive
                point = bz.points_primitive
            else:
                norms = bz.normals_invA
                point = bz.points_invA
            fvecs = [np.array([p, p + n]) for p, n in zip(point, norms)]
            lcol = Line3DCollection(fvecs)
            axs.add_collection3d(lcol)
        axs.set_xlim(left=xyz_min[0], right=xyz_max[0])
        axs.set_ylim(bottom=xyz_min[1], top=xyz_max[1])
        axs.set_zlim(bottom=xyz_min[2], top=xyz_max[2])
        if isinstance(Q, np.ndarray) and Q.ndim == 2 and Q.shape[1] == 3:
            axs.scatter(Q[:, 0], Q[:, 1], Q[:, 2])
        # axs.set_aspect('equal', 'box') # removed from newer Matplotlib
        # axs.auto_scale_xyz(1.,1.,1.) # supposed-workaround, probably need to set scaling based on figure size and view
        if show:
            pp.show()
        return axs

    def _make_poly_collection(
        verts,
        vpf,
        origin=None,
        color="b",
        edgecolor="k",
        linestyle="-",
        linewidth=1,
        alpha=0.5,
    ):
        # vpf lists the ordered vertices which make up each facet
        # for each facet, pick-out the vertices which define its polygon face
        patches = [np.array([verts[j, :] for j in i]) for i in vpf]
        # if an origin has been provided, add it to the patches
        if origin is not None and origin.ndim == 1 and origin.shape[0] == 3:
            for p in patches:
                p += origin
        # find the extent of the patches
        xyz_min = np.array([x.min() for x in np.vsplit(verts.transpose(), 3)])
        xyz_max = np.array([x.max() for x in np.vsplit(verts.transpose(), 3)])
        # plus some nice-for-plotting padding
        dif = xyz_max - xyz_min
        xyz_min -= dif / 20
        xyz_max += dif / 20
        # and create the collection of polygons in 3D
        collection = Poly3DCollection(
            patches,
            edgecolor=edgecolor,
            linestyle=linestyle,
            linewidth=linewidth,
            alpha=alpha,
        )
        # which requires that the face color be set after the fact
        collection.set_facecolor(color)
        return (collection, xyz_min, xyz_max)

    def __cube(p_0, p_1):
        """Return the patches of a cube bounded by points p_0 and p_1."""
        d_x = np.array((p_1[0] - p_0[0], 0, 0))
        d_y = np.array((0, p_1[1] - p_0[1], 0))
        d_z = np.array((0, 0, p_1[2] - p_0[2]))
        verts = p_0 + np.array(
            [
                d_x - d_x,  # 0 (000)
                d_x,  # 1 (100)
                d_x + d_y,  # 2 (110)
                d_y,  # 3 (010)
                d_z,  # 4 (001)
                d_z + d_x,  # 5 (101)
                d_z + d_x + d_y,  # 6 (111)
                d_z + d_y,
            ]
        )  # 7 (011)
        idx = np.array(
            [
                [0, 1, 2, 3],  # (000)-(100)-(110)-(010)
                [0, 1, 5, 4],  # (000)-(100)-(101)-(001)
                [0, 4, 7, 3],  # (000)-(001)-(011)-(010)
                [4, 5, 6, 7],  # (001)-(101)-(111)-(011)
                [6, 2, 1, 5],  # (111)-(110)-(100)-(101)
                [2, 6, 7, 3],
            ]
        )  # (110)-(111)-(011)-(010)
        patches = [verts[x] for x in idx]
        return patches

    def plot_polyhedron(poly, axs=None, setlims=True, show=True, **kwds):
        """Plot a single polyhedron.

        Parameters
        ----------
        poly : :py:class:`brille._brille.Polyhedron`
            Any object with both a ``vertices`` and ``vertices_per_face`` field
            could work with thie plotting function, however it is anticipated that a
            :py:class:`brille._brille.Polyhedron` will be provided.

        axs : :py:class:`matplotlib.axes.Axes`, optional
            The 3D axes in which to add the polyhedron facets. If ``None`` then
            :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
            axes.

        setlims : bool, optional
            Whether to change the limits of `axs` to match the limits of the extent
            of `poly.vertices`.

        show : bool, optional
            Whether to call `matplotlib.pyplot.show()` after adding the points to
            `axs`; this is mostly useful in non-interactive environments.

        origin : {:py:class:`numpy.ndarray`,tuple,list}, optional
            The origin of the plotting coordinate system, all drawn information is
            relative to this vector. Any 3-element object convertible to a
            :py:class:`numpy.ndarray` is valid input. Invalid input is replaced by
            the zero-vector.

        color : optional
            The face color of the drawn polygons.

        edgecolor : optional
            The edge color of the drawn polygons.

        linestyle : str, optional
            The edge line style of dranw polygons.

        linewidth : float, optional
            The edge line width of drawn polygons.

        alpha : float, optional
            The face alpha of drawn polygons.

        Returns
        -------
        :py:class:`matplotlib:axes:Axes`
            The value of `axs` after plotting.
        """

        # pylint: disable=no-member
        axs = _check_axes(axs)
        # the 1st Brillouin zone has on-face points equal to half the normals
        coll, xyz_min, xyz_max = _make_poly_collection(
            poly.vertices, poly.faces, **kwds
        )
        axs.add_collection3d(coll)
        if setlims:
            axs.set_xlim(left=xyz_min[0], right=xyz_max[0])
            axs.set_ylim(bottom=xyz_min[1], top=xyz_max[1])
            axs.set_zlim(bottom=xyz_min[2], top=xyz_max[2])
        if show:
            pp.show()
        return axs

    def plot_tetrahedron(verts, axs=None, show=True, **kwds):
        """Plot a single tetrahedron.

        Parameters
        ----------
        verts : :py:class:`numpy.ndarray`
            A :math:`4 \\times 3` array of the four vertices of the tetrahedron.

        axs : :py:class:`matplotlib.axes.Axes`, optional
            The 3D axes in which to add the polyhedron facets. If ``None`` then
            :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
            axes.

        show : bool, optional
            Whether to call `matplotlib.pyplot.show()` after adding the points to
            `axs`; this is mostly useful in non-interactive environments.

        origin : {:py:class:`numpy.ndarray`,tuple,list}, optional
            The origin of the plotting coordinate system, all drawn information is
            relative to this vector. Any 3-element object convertible to a
            :py:class:`numpy.ndarray` is valid input. Invalid input is replaced by
            the zero-vector.

        color : optional
            The face color of the drawn polygons.

        edgecolor : optional
            The edge color of the drawn polygons.

        linestyle : str, optional
            The edge line style of dranw polygons.

        linewidth : float, optional
            The edge line width of drawn polygons.

        alpha : float, optional
            The face alpha of drawn polygons.

        Returns
        -------
        :py:class:`matplotlib:axes:Axes`
            The value of `axs` after plotting.
        """
        if not (verts.ndim == 2 and verts.shape[0] == 4 and verts.shape[1] == 3):
            raise RuntimeError("Input are not the vertices of a tetrahedron")
        vpf = np.array([[0, 1, 2], [0, 3, 1], [3, 2, 1], [0, 2, 3]])
        pc, _, _ = _make_poly_collection(verts, vpf, **kwds)
        # Add the Poly3DCollection to existing or new axes:
        axs = _check_axes(axs)
        axs.add_collection3d(pc)
        if show:
            pp.show()
        return axs

    def plot_tetrahedra(allverts, tetidx, axs=None, **kwds):
        """Plot a number of tetrahedra.

        Parameters
        ----------
        allverts : :py:class:`numpy.ndarray`
            A :math:`(N \\ge 4) \\times 3` array of the vertices of all tetrahedra

        tetidx: :py:class:`numpy.ndarray`
            A :math:`M \\times 4` array of the indices of `allverts` which make up each of
            the :math:`M` tetrahedra to be plotted.
            The values of `tetidx` should obey the inequalities ``min(tetidx) ≥ 0``
            and ``max(tetidx) < N``.

        axs : :py:class:`matplotlib.axes.Axes`, optional
            The 3D axes in which to add the polyhedron facets. If ``None`` then
            :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
            axes.

        color : {arraylike, str, iterable}, optional
            The specified `color` will be used to produce a list of :math:`M` colors to
            use in plotting the :math:`M` tetrahedra.
            If `color` has three elements or is a `str` it is assumed to represent
            a single RGB value.
            In all cases the single or multiple colors provided in `color` are tiled
            into a list with at least :math:`M` elements before being truncated.
            If no `color` is provided, a list of all named colors known to
            :py:mod:`matplotlib.colors` is tiled.

        """
        if not (allverts.ndim == 2 and allverts.shape[1] == 3):
            raise RuntimeError("Vertices are not the correct shape")
        if isinstance(tetidx, list):
            tetidx = np.array(tetidx)
        if not (tetidx.ndim == 2 and tetidx.shape[1] == 4):
            raise RuntimeError("Tetrahedra indexes are not the correct shape")
        colours = make_colours(tetidx.shape[0], **kwds)
        # we want to ensure all tetrahedra end up in the same set of axes
        axs = _check_axes(axs)
        for tet, colour in zip(tetidx, colours):
            plot_tetrahedron(allverts[tet], color=colour, **kwds)

    def make_colours(n, color=None, **kwds):
        if color is None:
            from matplotlib.colors import get_named_colors_mapping

            color = get_named_colors_mapping().keys()

        from collections.abc import Iterable

        if isinstance(color, Iterable):
            color = list(color)
        if isinstance(color, str) or (
            isinstance(color, (list, tuple)) and len(color) == 3
        ):
            color = [color]
        if not isinstance(color, np.ndarray):
            color = np.array(color)
        if color.shape[0] < n:
            color = np.tile(color, 1 + n // color.shape[0])
        return color[0:n]



# Build and test the CMake project in a manylinux container using a persistent mounted folder

ENTRYPOINT = """#!/bin/bash
set -eu
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# ! This file is auto-generated and will be overwritten !
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

function echo_run {{
    echo "-------------------------------------------------"
    echo "$" "$@"
    echo "-------------------------------------------------"
    "$@"
}}

export PATH={path}:$PATH

# Install HDF5 development files
# echo_run yum install -y hdf5-devel

# Ensure the python interpreter has setuptools_scm for versioning 
echo_run {python} -m pip install setuptools_scm conan

# Configure the project, specifying which python interpreter and library to use
echo_run export CONAN_USER_HOME=/conan
echo_run cmake -S /source -B /build -DPYTHON_EXECUTABLE={python}

# Build the testing project
echo_run cmake --build /build --target tester --parallel {count}

# Run the tests
echo_run ctest --test-dir /build --parallel {count} --rerun-failed --output-on-failure
"""


def main():
    from config import get_client, get_image, get_volumes, get_folder, write_entrypoint
    from python_on_whales import exceptions

    client = get_client()
    image = get_image(client)
    folder = get_folder("cmake_build")
    volumes = get_volumes(client, {"build": folder, "conan": None})

    write_entrypoint(ENTRYPOINT, folder)
    try:
        result = client.run(
            image, ["sh", "/build/entrypoint.sh"], volumes=volumes, tty=True
        )
        print(result)
    except exceptions.DockerException as ex:
        raise RuntimeError(ex)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Run tests of the interpolation functionality."""

import unittest
import numpy as np


def sqwfunc_ones(Q):
    """S(Q,W) function that is all ones."""
    return 1.0 + 0 * (Q[:, 0] + Q[:, 1] + Q[:, 2])


def sqwfunc_x(Q):
    """S(Q,W) function that returns Qx."""
    return Q[:, 0]


def sqwfunc_y(Q):
    """S(Q,W) function that returns Qy."""
    return Q[:, 1]


def sqwfunc_z(Q):
    """S(Q,W) function that returns Qz."""
    return Q[:, 2]


def sqwfunc_xy(Q):
    """S(Q,W) function that returns Qx+Qy."""
    return Q[:, 0] + Q[:, 1]


def sqwfunc_xyz(Q):
    """S(Q,W) function that returns Qx+Qy+Qz."""
    return Q[:, 0] + Q[:, 1] + Q[:, 2]


def vecfun_ident(Q):
    """S(Q,W) function that returns vector Q."""
    return Q


def vecfun_rotz(Q, θ=0):
    """S(Q,W) function that returns vector Q rotated around the z-axis."""
    x = Q[:, [0]]
    y = Q[:, [1]]
    z = Q[:, [2]]
    c = np.cos(θ)
    s = np.sin(θ)
    return np.concatenate((x * c - y * s, x * s + y * c, z), axis=1)


def vecfun_rotx(Q, θ=0):
    """S(Q,W) function that returns vector Q rotated around the x-axis."""
    x = Q[:, [0]]
    y = Q[:, [1]]
    z = Q[:, [2]]
    c = np.cos(θ)
    s = np.sin(θ)
    return np.concatenate((x, y * c - z * s, y * s + z * c), axis=1)


def matfun_ident(Q):
    """S(Q,W) function that returns a flattend matrix with Q along its diagonal."""
    sh = Q.shape
    z = np.ndarray((sh[0], sh[1] * sh[1]), dtype=Q.dtype)
    for i in range(sh[0]):
        z[i, :] = np.diag(Q[i, :]).flatten()
    return z


def modefun(Q):
    """S(Q,W) function that returns multiple modes."""
    sh = Q.shape
    modes = np.ndarray((sh[0], sh[1], sh[1], 3), dtype=Q.dtype)
    for i in range(sh[0]):
        modes[i, :, :, 0] = np.diag(Q[i, :])
        modes[i, 0, :, 1] = Q[i, :]
        modes[i, :, 0, 2] = -Q[i, :]
        # modes[i, sh[1]-1, :, 3] = 1-Q[i, :]
        # modes[i, :, sh[1]-1, 4] = 1+Q[i, :]
    return modes


def complex_scalar(Q):
    """S(Q,W) function that returns a complex scalar number."""
    return (Q[:, 0] - Q[:, 1]) + (Q[:, 2] + Q[:, 1]) * 1j


def setup_grid(iscomplex=False, halfN=(2, 2, 2)):
    """Create a grid object for interpolating."""
    from brille import Lattice, BrillouinZone, BZTrellisQcc, BZTrellisQdd

    lat = Lattice(((1, 1, 1), (90, 90, 90)), real_space=False)
    bz = BrillouinZone(lat)
    max_volume = lat.volume_star / (8 * np.prod(halfN))
    if iscomplex:
        bzg = BZTrellisQcc(bz, max_volume)
    else:
        bzg = BZTrellisQdd(bz, max_volume)
    return bzg


def define_Q_points(rand=False, N=10):
    """Define a number of Q points for use in inerpolating."""
    if rand:
        # the tests only work if the Q points are already within the first BZ
        Q = np.random.rand(N, 3) - 0.5  # (-0.5,0.5)
    else:
        Q = np.array([[0, 0, 0], [0.1, 0, 0], [0, 0.1, 0], [0, 0, 0.1]], dtype="double")
    return Q


def fe_dispersion(Q, p=(-16, 0.01)):
    """Calculate the dispersion relationship for spinwaves in iron."""
    J = p[0]  # exchange, meV
    d = p[1]  # anisotropy
    # the dispersion relationship:
    q_dep = np.cos(np.pi * Q[:, 0]) * np.cos(np.pi * Q[:, 1]) * np.cos(np.pi * Q[:, 2])
    w = d + 8 * J * (1 - q_dep)
    return w


def fe_analytic(Q, E, p):
    """Calculate the analytic function for S(Q,ω) for spinwaves in iron."""
    g = p[2]  # mode-lifetime, meV
    T = p[3]  # Temperature, K
    A = p[4]  # scale-factor
    w = fe_dispersion(Q, p)
    S = (
        A
        / np.pi
        * (E / 1 - np.exp(-11.602 * E / T))
        * (4 * g * w)
        / ((E**2 - w**2) ** 2 + 4 * (g * E) ** 2)
    )
    return S


class Interpolate(unittest.TestCase):
    """Class to perform unit tests related to interpolation."""

    def test_a_norm(self):
        """Test interpolation normalisation."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_ones(bzg.rlu), (1,), bzg.rlu, (0, 3))
        interpolated_ones, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.allclose(interpolated_ones, 1.0))

    def test_b_x(self):
        """Test with data as Qx."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_x(bzg.rlu), (1,), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(np.squeeze(intres), sqwfunc_x(Qi)).all())

    def test_b_y(self):
        """Test with data as Qy."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_y(bzg.rlu), (1,), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(np.squeeze(intres), sqwfunc_y(Qi)).all())

    def test_b_z(self):
        """Test with data as Qz."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_z(bzg.rlu), (1,), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(np.squeeze(intres), sqwfunc_z(Qi)).all())

    def test_c_xy(self):
        """Test with data as Qx+Qy."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_xy(bzg.rlu), (1,), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(np.squeeze(intres), sqwfunc_xy(Qi)).all())

    def test_c_xyz(self):
        """Test with data as Qx+Qy+Qz."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_xyz(bzg.rlu), (1,), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(np.squeeze(intres), sqwfunc_xyz(Qi)).all())

    def test_d_vec_ident(self):
        """Test with data as vector Q."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(vecfun_ident(bzg.rlu), (0, 3), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi).all())

    def test_d_vec_rotx(self):
        """Test with data as vector Q rotated about x."""
        bzg = setup_grid()
        Qi = define_Q_points()
        ang = np.pi / 3
        bzg.fill(vecfun_rotx(bzg.rlu, ang), (0, 3), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        antres = vecfun_rotx(Qi, ang)
        self.assertTrue(np.isclose(intres, antres).all())

    def test_d_vec_rotz(self):
        """Test with data as vector Q rotated about z."""
        bzg = setup_grid()
        Qi = define_Q_points()
        ang = 3 * np.pi / 5
        bzg.fill(vecfun_rotz(bzg.rlu, ang), (0, 3), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        antres = vecfun_rotz(Qi, ang)
        self.assertTrue(np.isclose(intres, antres).all())

    def test_e_mat_ident(self):
        """Test with data as matrix Q."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(matfun_ident(bzg.rlu), (0, 0, 9), bzg.rlu, (0, 3))
        intres, _ = bzg.interpolate_at(Qi)
        antres = matfun_ident(Qi)
        self.assertTrue(np.isclose(intres, antres).all())

    def test_f_complex_scalar(self):
        """Test with data as complex-valued scalars."""
        bzg = setup_grid(iscomplex=True)
        Qi = define_Q_points()
        complexQ = complex_scalar(bzg.rlu)
        bzg.fill(complexQ, (1,), complexQ, (1,))
        # complex eigenvalues have fixed phase while complex eigenvectors
        # are allowed arbitrary phase.
        val, vec = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(np.squeeze(val), complex_scalar(Qi)).all())

    # def test_g_big_grid(self):
    #     """Test with a large grid."""
    #     bzg = setup_grid(halfN=(20, 20, 20))
    #     Qi = define_Q_points(rand=True, N=100000)
    #     bzg.fill(matfun_ident(bzg.rlu))
    #     intres = bzg.interpolate_at(Qi, True, True, 10)
    #     antres = matfun_ident(Qi)
    #     self.assertTrue(np.isclose(intres, antres).all())

    def test_i_iron_self_consistency(self):
        """Test with data as iron spinwaves, but test only *at* grid points."""
        from brille import Lattice, BrillouinZone, BZTrellisQdc

        lat = Lattice(
            ((2.87, 2.87, 2.87), np.pi / 2 * np.array((1, 1, 1))), spacegroup="Im-3m"
        )
        bz = BrillouinZone(lat)  # constructs an irreducible Bz by default
        bzg = BZTrellisQdc(bz, 0.125)
        Q = bzg.rlu
        bzg.fill(fe_dispersion(Q), (1,), bzg.rlu, (0, 3))
        # The irreducible interpolation must be used here
        # since the Brillouin zone is an *irreducible* Brillouin zone!!
        intres, _ = bzg.ir_interpolate_at(Q, False, False)
        antres = fe_dispersion(Q)
        self.assertTrue(np.isclose(np.squeeze(intres), antres).all())

    def test_j_stored_data_references(self):
        """Ensure that the reference count for stored data is incremented"""
        from sys import getrefcount as r

        bzg = setup_grid()
        val = sqwfunc_xyz(bzg.rlu)
        vec = vecfun_rotz(bzg.rlu, np.pi / 180 * 40)
        # Store the value and vector array reference counts before filling
        valr, vecr = r(val), r(vec)
        # fill the data into the grid -- this increases each array refcount
        bzg.fill(val, (1,), vec, (0, 3))
        self.assertTrue(valr + 1 == r(val))
        self.assertTrue(vecr + 1 == r(vec))


if __name__ == "__main__":
    unittest.main()



# The compiled module is installed at the base of the package directory structure such that its properties
# can be included here. If the module is loaded, however, without being installed first these import statements will
# raise a ModuleNotFoundError, which prevents automated pre-installation testing from taking place.
from importlib.util import find_spec

bound_spec = find_spec("brille._brille")  # won't this cause a circular import?
if bound_spec is not None:
    try:
        from ._brille import (
            __version__,
            version,
            AngleUnit,
            LengthUnit,
            Bravais,
            RotatesLike,
            NodeType,
            real_space_tolerance,
            reciprocal_space_tolerance,
            ApproxConfig,
            Basis,
            HallSymbol,
            Spacegroup,
            Pointgroup,
            Symmetry,
            PointSymmetry,
            PrimitiveTransform,
            BrillouinZone,
            Polyhedron,
            LPolyhedron,
            SortingStatus,
            BZTrellisQcc,
            BZTrellisQdc,
            BZTrellisQdd,
            BZMeshQcc,
            BZMeshQdc,
            BZMeshQdd,
            BZNestQcc,
            BZNestQdc,
            BZNestQdd,
        )

        # Store the grid types for use in, e.g., the plotting routines
        __grid_types__ = (
            BZTrellisQcc,
            BZTrellisQdc,
            BZTrellisQdd,
            BZMeshQcc,
            BZMeshQdc,
            BZMeshQdd,
            BZNestQcc,
            BZNestQdc,
            BZNestQdd,
        )

        __all__ = [
            "__version__",
            "version",
            "AngleUnit",
            "LengthUnit",
            "Bravais",
            "RotatesLike",
            "NodeType",
            "real_space_tolerance",
            "reciprocal_space_tolerance",
            "ApproxConfig",
            "Basis",
            "HallSymbol",
            "Spacegroup",
            "Pointgroup",
            "Symmetry",
            "PointSymmetry",
            "PrimitiveTransform",
            "BrillouinZone",
            "Polyhedron",
            "LPolyhedron",
            "SortingStatus",
            "BZTrellisQcc",
            "BZTrellisQdc",
            "BZTrellisQdd",
            "BZMeshQcc",
            "BZMeshQdc",
            "BZMeshQdd",
            "BZNestQcc",
            "BZNestQdc",
            "BZNestQdd",
            "__grid_types__",
        ]
    except ImportError as imp_error:
        import platform

        if platform.system == "Windows" and "DLL" in imp_error.msg:
            msg = "You may be missing the latest Visual C++ redistributable package"
            msg += " install it from Microsoft @ https://support.microsoft.com/en-us/help/2977003"
            msg += " before trying to import brille again"
            print(msg)
        raise imp_error
else:
    __version__ = "0.0.0"
    __all__ = ["__version__"]



#!/usr/bin/env python3
import unittest
import numpy as np

import brille as br_py
from brille import _brille as br_mod

if br_mod != br_py:
    print(
        f"Loaded binary module {br_mod.__file__} and python module {br_py.__file__} differ"
    )


def get_latvec(lens, angs):
    # Calculates the lattice vectors after the convention of self.spglib
    xhat = np.array([1, 0, 0])
    yhat = np.array([np.cos(angs[2]), np.sin(angs[2]), 0])
    vol = np.sqrt(1 - np.sum(np.cos(angs) ** 2) + 2 * np.prod(np.cos(angs)))
    zhat = np.array(
        [
            np.cos(angs[1]),
            (np.cos(angs[0]) - np.cos(angs[1]) * np.cos(angs[2])) / np.sin(angs[2]),
            vol / np.sin(angs[2]),
        ]
    )
    return np.array([lens[0] * xhat, lens[1] * yhat, lens[2] * zhat])


class UtilsTestBZ(unittest.TestCase):
    # Tests create_bz routines
    a, b, c, alpha, beta, gamma = 4.0, 4.0, 5.0, np.pi / 2, np.pi / 2, 2 * np.pi / 3
    spg = "P 6"

    def check_lattice(self, bz, vals=None):
        lattice = bz.lattice
        if not vals:
            vals = [self.a, self.b, self.c, self.alpha, self.beta, self.gamma]
        self.assertAlmostEqual(lattice.a, vals[0])
        self.assertAlmostEqual(lattice.b, vals[1])
        self.assertAlmostEqual(lattice.c, vals[2])
        self.assertAlmostEqual(lattice.alpha, vals[3])
        self.assertAlmostEqual(lattice.beta, vals[4])
        self.assertAlmostEqual(lattice.gamma, vals[5])

    def test_separate_latpars(self):
        bz = br_py.utils.create_bz(
            self.a, self.b, self.c, self.alpha, self.beta, self.gamma, self.spg
        )
        self.check_lattice(bz)
        # Again with keywords
        bz = br_py.utils.create_bz(
            a=self.a,
            b=self.b,
            c=self.c,
            alpha=self.alpha,
            beta=self.beta,
            gamma=self.gamma,
            spacegroup=self.spg,
        )
        self.check_lattice(bz)

    def test_separate_latpars_degrees(self):
        alf, bet, gam = np.array([self.alpha, self.beta, self.gamma]) / np.pi * 180
        bz = br_py.utils.create_bz(self.a, self.b, self.c, alf, bet, gam, self.spg)
        self.check_lattice(bz)

    def test_lengths_angles(self):
        bz = br_py.utils.create_bz(
            [self.a, self.b, self.c], [self.alpha, self.beta, self.gamma], self.spg
        )
        self.check_lattice(bz)
        # With keywords
        bz = br_py.utils.create_bz(
            lens=[self.a, self.b, self.c],
            angs=[self.alpha, self.beta, self.gamma],
            spacegroup=self.spg,
        )
        self.check_lattice(bz)

    def test_lengths_angles_degrees(self):
        bz = br_py.utils.create_bz(
            [self.a, self.b, self.c],
            np.array([self.alpha, self.beta, self.gamma]) / np.pi * 180,
            self.spg,
        )
        self.check_lattice(bz)

    def test_lattice_vectors(self):
        latvec = get_latvec(
            [self.a, self.b, self.c], [self.alpha, self.beta, self.gamma]
        )
        bz = br_py.utils.create_bz(latvec, self.spg)
        self.check_lattice(bz)

    def test_invalid_input_wrong_spg(self):
        # Tests incorrect spacegroup for lattice parameters - cannot find IR wedge
        with self.assertRaises(RuntimeError):
            bz = br_py.utils.create_bz(
                [4, 5, 6], [90, 90, 90], "P 6", snap_to_symmetry=False
            )
        bz = br_py.utils.create_bz(
            [4, 5, 6], [90, 90, 90], "P 6", wedge_search=False, snap_to_symmetry=False
        )
        self.check_lattice(bz, [4, 5, 6] + [np.pi / 2] * 3)

    def test_invalid_input_negative_squared_vol(self):
        # Tests for physically impossible lattice
        with self.assertRaises(ValueError):
            br_py.utils.create_bz([4, 5, 6], [30, 30, 90], "P 6")

    def test_no_spacegroup(self):
        # Tests if no spacegroup is given
        with self.assertRaises(ValueError):
            br_py.utils.create_bz([4, 5, 6], [90, 90, 90])

    # 'Hall number' specification removed in v0.6.0
    # def test_spacegroup_number(self):
    #     # Tests for a spacegroup / Hall number instead of a Hall symbol
    #     bz = br_py.utils.create_bz(self.a, self.b, self.c, self.alpha, self.beta, self.gamma, 1)
    #     self.check_lattice(bz)

    def test_invalid_spacegroups(self):
        # Tests for invalid spacegroups given
        with self.assertRaises(ValueError):
            # Must be a string
            br_py.utils.create_bz([4, 5, 6], [90, 90, 90], [])
        with self.assertRaises(ValueError):
            # Must be a string
            br_py.utils.create_bz([4, 5, 6], [90, 90, 90], -1)

    def test_using_reciprocal(self):
        # Test with reciprocal lattice
        latvec = get_latvec(
            [self.a, self.b, self.c], [self.alpha, self.beta, self.gamma]
        )
        bz = br_py.utils.create_bz(
            2 * np.pi * np.linalg.inv(latvec).T, self.spg, is_reciprocal=True
        )
        self.check_lattice(bz)


class UtilsTestGrid(unittest.TestCase):
    # Tests create_grid routines
    bz = br_py.utils.create_bz(4, 4, 6, 90, 90, 110, "P 2")

    def check_type(self, expected_type, *args, **kwargs):
        obj = br_py.utils.create_grid(self.bz, *args, **kwargs)
        self.assertTrue(isinstance(obj, expected_type))

    def test_expected_types(self):
        # Checks the input arguments results in the correct grid types
        self.check_type(br_mod.BZTrellisQdd, node_volume_fraction=0.1)
        self.check_type(
            br_mod.BZTrellisQdc, node_volume_fraction=0.1, complex_vectors=True
        )
        self.check_type(
            br_mod.BZTrellisQcc,
            node_volume_fraction=0.1,
            complex_vectors=True,
            complex_values=True,
        )
        self.check_type(br_mod.BZMeshQdd, mesh=True)
        self.check_type(br_mod.BZMeshQdc, mesh=True, complex_vectors=True)
        self.check_type(
            br_mod.BZMeshQcc, mesh=True, complex_vectors=True, complex_values=True
        )
        self.check_type(br_mod.BZNestQdd, nest=True, max_volume=0.1)
        self.check_type(
            br_mod.BZNestQdc, nest=True, max_volume=0.1, complex_vectors=True
        )
        self.check_type(
            br_mod.BZNestQcc,
            nest=True,
            max_volume=0.1,
            complex_vectors=True,
            complex_values=True,
        )
        self.check_type(br_mod.BZNestQdd, nest=True, number_density=10.0)
        self.check_type(
            br_mod.BZNestQdc, nest=True, number_density=10.0, complex_vectors=True
        )
        self.check_type(
            br_mod.BZNestQcc,
            nest=True,
            number_density=10.0,
            complex_vectors=True,
            complex_values=True,
        )

    def test_bad_input(self):
        # Checks for invalid input
        with self.assertRaises(ValueError):
            # Specifying both mesh and nest [cannot decide]
            br_py.utils.create_grid(self.bz, mesh=True, nest=True)
        with self.assertRaises(ValueError):
            # Parameters for BZNestQ* without nest=True
            br_py.utils.create_grid(self.bz, max_volume=0.1)
        with self.assertRaises(ValueError):
            # Missing arguments `max_volume` or `number_density`
            br_py.utils.create_grid(self.bz, nest=True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Run tests of Brillouin zone creation."""

import unittest


def get_local_JSON(filename):
    from pathlib import Path
    import json

    with open(Path(__file__).parent / filename) as file:
        data = json.load(file)
    return data


def get_aflow_lattices():
    return get_local_JSON("aflow_lattices.json")


def n2chr(n):
    return chr(48 + n) if n < 10 else chr(65 - 10 + n)


class AflowTest(unittest.TestCase):
    def test_aflow_crystaldatabase(self):
        from numpy import zeros, isclose
        from brille import Lattice, BrillouinZone, PointSymmetry

        tested = 0
        failed = 0
        errored = 0
        failed_afl = []
        failed_ratio = []
        errored_afl = []
        errored_arg = []
        hall_groups_passed = zeros(530, dtype="int")
        hall_groups_failed = zeros(530, dtype="int")
        for afl in get_aflow_lattices():
            # afl == [hall_number, basis_vector_lengths, basis_vector_angles, Hall_symbol]
            lat = Lattice((afl[1], afl[2]), spacegroup=afl[3])
            tested += 1
            try:
                bz = BrillouinZone(lat)
                vol_bz = bz.polyhedron.volume
                vol_ir = bz.ir_polyhedron.volume
                if not isclose(vol_ir, vol_bz / PointSymmetry(afl[0]).size):
                    failed += 1
                    failed_afl.append(afl)
                    failed_ratio.append(vol_ir / vol_bz * PointSymmetry(afl[0]).size)
                    hall_groups_failed[afl[0] - 1] += 1
                else:
                    hall_groups_passed[afl[0] - 1] += 1
            except Exception as err:
                errored += 1
                errored_afl.append(afl)
                errored_arg.append(err.args)
        if failed > 0:
            print(
                "\nFailed to find correct irreducible Brillouin zone for",
                failed,
                "out of",
                tested,
                "lattices",
            )
            for file, rat in zip(failed_afl, failed_ratio):
                print(file, rat)
        if errored > 0:
            print("\nException raised for", errored, "out of", tested, "lattices")
            for file, arg in zip(errored_afl, errored_arg):
                print(file, arg)
        print(
            "\nHall groups passed (total =",
            hall_groups_passed.sum(),
            "of",
            tested,
            "tested)",
        )
        encoded_hgp = [n2chr(x) for x in hall_groups_passed]
        for x in [encoded_hgp[i * 53 : (i + 1) * 53] for i in range(10)]:
            print("".join(x))

        self.assertTrue(failed == 0)
        self.assertTrue(errored == 0)


if __name__ == "__main__":
    unittest.main()



# Build the documentation pages in a reproducible environment

ENTRYPOINT = r"""#!/bin/sh
set -eu
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# ! This file is auto-generated and will be overwritten !
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

function echo_run {{
    echo "-------------------------------------------------"
    echo "$" "$@"
    echo "-------------------------------------------------"
    "$@"
}}

export PATH={path}:$PATH
export CONAN_USER_HOME=/conan

# Configure the project, specifying which python interpreter and library to use
echo_run echo $CONAN_USER_HOME
echo_run ls -la $CONAN_USER_HOME
echo_run {python} -m pip wheel -w /build --no-deps /source

echo_run find /build -type f -iname "*-linux*.whl" -exec sh -c "auditwheel repair '{{}}' -w \$(dirname '{{}}')" \;

"""

PAGES_ENTRYPOINT = """#!/bin/sh
set -eu
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# ! This file is auto-generated and will be overwritten !
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

function echo_run {{
    echo "-------------------------------------------------"
    echo "$" "$@"
    echo "-------------------------------------------------"
    "$@"
}}

sphinx_doctree=/conan/.doctree
echo_run mkdir -p $sphinx_doctree

html_dir=`mktemp -d`

echo_run git config --global --add safe.directory /source
echo_run cd /source

echo_run ls /wheelhouse
echo_run python3 -m pip install /wheelhouse/{wheel}

echo_run mkdir -p /source/docs/_build/doxygenxml

echo_run sphinx-build -b html /source/docs $html_dir -E -d $sphinx_doctree

echo_run mkdir -p /build/pages
echo_run rsync -a --delete "${{html_dir}}/" /build/pages

"""


def find_wheel(p):
    return [x.name for x in p.iterdir() if x.is_file() and "cp39-musllinux" in str(x)]


def main(build_wheel, wheelhouse, output):
    from config import get_client, get_image, get_volumes, get_folder, write_entrypoint
    from python_on_whales import exceptions

    client = get_client()
    folder = get_folder(wheelhouse)

    if not build_wheel:
        wheel = find_wheel(folder)
    else:
        wheel = []

    if len(wheel) != 1:
        # Build the musllinux wheel
        image = get_image(client, "quay.io/pypa/musllinux_1_1_x86_64")
        volumes = get_volumes(client, {"build": folder, "conan": None})
        print(volumes)
        write_entrypoint(ENTRYPOINT, folder, "cp39-cp39")
        try:
            result = client.run(
                image, ["sh", "/build/entrypoint.sh"], volumes=volumes, tty=True
            )
            print(result)
        except exceptions.DockerException as ex:
            raise RuntimeError(ex)
        wheel = find_wheel(folder)

    # Use the built wheel to build the documentation
    if len(wheel) != 1:
        raise RuntimeError(
            f"Failed to find single musllinux wheel; found {wheel} instead"
        )
    image = get_image(client, "docker.io/brille/docact:v5.0.0")
    volumes = get_volumes(
        client, {"wheelhouse": folder, "build": get_folder(output), "conan": None}
    )
    write_entrypoint(PAGES_ENTRYPOINT, folder, sub={"wheel": wheel[0]})
    try:
        result = client.run(
            image, ["sh", "/wheelhouse/entrypoint.sh"], volumes=volumes, tty=True
        )
    except exceptions.DockerException:
        raise RuntimeError()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Build image-based Sphinx documentation locally for brille"
    )
    parser.add_argument(
        "-s",
        "--skip-wheel",
        action="store_true",
        help="Skip building the musl linux wheel if it already exists in WHEELHOUSE",
    )
    parser.add_argument(
        "-w",
        "--wheelhouse",
        type=str,
        default="wheelhouse",
        help='Wheelhouse directory, default="wheelhouse"',
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default="pages_build",
        help='Output HTML paged directory, default="pages_build"',
    )
    args = parser.parse_args()

    main(not args.skip_wheel, args.wheelhouse, args.output)



# Build and test the python module using CIBuildWheel

ENTRYPOINT = r"""#!/bin/bash
set -eu
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# ! This file is auto-generated and will be overwritten !
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

function echo_run {{
    echo "-------------------------------------------------"
    echo "$" "$@"
    echo "-------------------------------------------------"
    "$@"
}}

export PATH={path}:$PATH
export CONAN_USER_HOME=/conan

# Configure the project, specifying which python interpreter and library to use
echo_run {python} -m pip wheel -w /build --no-deps /source

echo_run find /build -type f -iname "*-linux*.whl" -exec sh -c "auditwheel repair '{{}}' -w \$(dirname '{{}}')" \;

"""


def main(variant=None):
    from config import get_client, get_image, get_volumes, get_folder, write_entrypoint
    from python_on_whales import exceptions

    client = get_client()
    image = get_image(client)
    folder = get_folder("wheelhouse")
    volumes = get_volumes(client, {"build": folder, "conan": None})

    write_entrypoint(ENTRYPOINT, folder, variant=variant)
    try:
        result = client.run(
            image, ["sh", "/build/entrypoint.sh"], volumes=volumes, tty=True
        )
        print(result)
    except exceptions.DockerException as ex:
        raise RuntimeError(ex)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("-p", "--python", type=str, default="cp312-cp312")
    args = parser.parse_args()
    main(variant=args.python)



import os
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as pp
import brille.plotting as bp
from brilleu.brilleu import getBrillEuObj

images_dir = os.path.join(os.path.dirname(__file__), 'images')


def dirsavefig(dirname, filename, transparent=True, **kwds):
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    pp.savefig(os.path.join(dirname, filename), transparent=transparent, **kwds)


def centre_to_corner(X):
    dX = np.diff(X, axis=0) / 2
    X = np.concatenate((
        X[:-1] - dX,
        X[np.newaxis, -1] - dX[np.newaxis, -1],
        X[np.newaxis, -1] + dX[np.newaxis, -1]), axis=0)
    dY = np.diff(X, axis=1) / 2
    X = np.concatenate((
        X[:, :-1] - dY,
        X[:, np.newaxis, -1] - dY[:, np.newaxis, -1],
        X[:, np.newaxis, -1] + dY[:, np.newaxis, -1]), axis=1)
    return X


# use_c=False ensures the Euphonic C module is *not* used.
# but *need* to use the brille C++ module, and parallel=True ensures we do so with OpenMP
scattering_lengths = {'Na': 3.63, 'Cl': 9.577}  # in fm
nacl = getBrillEuObj('NaCl', max_volume=1e-5, parallel=True, sort=True, scattering_lengths=scattering_lengths)

# Define a path through reciprocal space from (000) to (123)
xi = np.linspace(0, 1, 100)
qhkl = np.vstack((xi, 2 * xi, 3 * xi)).T
# Interpolate the mode energies at qhkl
w_q = nacl.w_q(qhkl)

bz = nacl.grid.BrillouinZone
# plot the first Brillouin zones for (000), (111), (022), and (113) and path
fig = pp.figure(figsize=pp.figaspect(2))
ax = fig.add_axes([0, 0, 1, 1], projection='3d')
ax.view_init(5, 300)
for g in [(0, 0, 0), (1, 1, 1), (0, 2, 2), (1, 1, 3)]:
    bp.plot(bz, units='rlu', irreducible=False, origin=g, edgecolor='gray', alpha=0, axs=ax, show=mpl.is_interactive())
bp.plot(qhkl, axs=ax, show=mpl.is_interactive())
pp.setp(ax, 'xlim', [-1, 2], 'ylim', [-1, 2], 'zlim', [-1, 4])
pp.setp(ax, 'xlabel', 'h', 'ylabel', 'k', 'zlabel', 'l')
bbox = fig.bbox_inches.from_bounds(0.2, 0.8, 4.4, 5.7)
dirsavefig(images_dir, 'nacl_123_path.png', bbox_inches=bbox)

# plot the Brillouin zone, path, irreducible path, and dispersion
irqhkl, G, rot, invrot = nacl.grid.BrillouinZone.ir_moveinto(qhkl)
fig = pp.figure(figsize=pp.figaspect(0.5))
ax0 = fig.add_subplot(1, 2, 1, projection='3d')
bp.plot(bz, units='rlu', axs=ax0, Q=irqhkl, show=mpl.is_interactive())
pp.setp(ax0, 'xlabel', 'h', 'ylabel', 'k', 'zlabel', 'l')

ax1 = fig.add_subplot(1, 2, 2)
for y in w_q.T:
    ax1.plot(xi, y)
ax1.set_xlabel(r'$\xi$ in $\mathbf{Q}=(\xi\,2\xi\,3\xi)$')
ax1.set_ylabel(r'$\omega_i(\mathbf{Q})$ / meV')

fig.subplots_adjust(left=0, right=0.99, bottom=0.1, top=1)
dirsavefig(images_dir, 'nacl_123_disp.png')

# determine and plot the inelastic neutron scattering intesity along the path
mu, nu = np.mgrid[0:1:50j, 0.5:30:60j]
ξ = mu.reshape(mu.size, 1)
En = nu.reshape(nu.size, 1)
Qhkl = np.concatenate((ξ, 2 * ξ, 3 * ξ), axis=1)

brSQE = nacl(Qhkl, En, resfun='sho', param=1., T=1.).reshape(mu.shape)
euSQE = nacl(Qhkl, En, resfun='sho', param=1., T=1., interpolate=False).reshape(mu.shape)

mu = centre_to_corner(mu)
nu = centre_to_corner(nu)
fig = pp.figure(figsize=pp.figaspect(0.5))
ax2 = fig.add_axes([0.07, 0.1, 0.4, 0.8])
ax2.pcolormesh(mu, nu, brSQE, shading='flat')
ax2.set_xlabel(r'$\xi$ in $\mathbf{Q}=(\xi\,2\xi\,3\xi)$')
ax2.set_ylabel(r'$E$/meV')
ax2.set_title('brille')

ax3 = fig.add_axes([0.55, 0.1, 0.4, 0.8])
ax3.pcolormesh(mu, nu, euSQE, shading='flat')
ax3.set_xlabel(r'$\xi$ in $\mathbf{Q}=(\xi\,2\xi\,3\xi)$')
ax3.set_ylabel(r'$E$/meV')
ax3.set_title('euphonic')

dirsavefig(images_dir, 'nacl_123_sqw.png')



from conan import ConanFile
from conan.tools.cmake import cmake_layout


class BrilleRecipe(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
    generators = "CMakeToolchain", "CMakeDeps"
    default_options = {
        "hdf5/*:shared": False,
        "hdf5/*:hl": False,
        "hdf5/*:with_zlib": False,
        "highfive/*:with_boost": False,
        "highfive/*:with_eigen": False,
        "highfive/*:with_xtensor": False,
        "highfive/*:with_opencv": False,
    }

    def requirements(self):
        self.requires("pybind11/2.13.1")
        # self.requires("hdf5/1.12.0")
        self.requires("catch2/3.6.0")
        self.requires("highfive/2.9.0")
        if self.settings.os == "Macos":
            self.requires("llvm-openmp/20.1.6")

    def configure(self):
        if self.settings.os == "Macos":
            self.options["llvm-openmp/*"].shared = True

    def layout(self):
        cmake_layout(self)



# Copyright © 2020,2021 Duc Le <duc.le@stfc.ac.uk>
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""
Utilities for ``brille``
------------------------

These functions help to construct lattices and grids for brillouin zone
interpolation.

.. code-block:: python

  import numpy as np
  from brille.utils import create_bz, create_grid

  def dispersion(q):
    energies = np.sum(np.cos(np.pi * q), axis=1)
    eigenvectors = np.sin(np.pi * q)
    return energies, eigenvectors

  bz = create_bz([4.1, 4.1, 4.1], [90, 90, 90], spacegroup='F m -3 m')
  grid = create_grid(bz, node_volume_fraction=1e-6)
  energies, eigenvectors = dispersion(grid.rlu)
  n_energies = 1     # Only one mode
  n_eigenvectors = 3 # Three values per q
  rotateslike = 0    # See note below
  grid.fill(energies, [n_energies, 0, 0, rotateslike],
            eigenvectors, [n_eigenvectors, 0, 0, rotateslike])

  interp_en, interp_ev = grid.ir_interpolate_at(np.random.rand(1000, 3))

Note
----
The ``rotateslike`` enumeration is given in :py:meth:`~brille._brille.BZMeshQdc.fill`
and describes how the eigenvalues / eigenvectors should be treated on application of a
symmetry operation. The *gamma* option (``rotateslike=3``) should be used for
phonon eigenvectors.


.. currentmodule:: brille.utils

.. autosummary::
    :toctree: _generate
"""

import numpy as np


def create_bz(
    *args,
    is_reciprocal=False,
    use_primitive=True,
    search_length=1,
    time_reversal_symmetry=False,
    wedge_search=True,
    snap_to_symmetry=True,
    **kwargs,
):
    """
    Construct a BrillouinZone object.

    Parameters
    ----------
    a, b, c : float
        Lattice parameters as separate floating point values
    lens : (3,) :py:class:`numpy.ndarray` or list
        Lattice parameters as a 3-element array or list
    alpha, beta, gamma : float
        Lattice angles in degrees or radians as separate floating point values
        Brille tries to determine if the input is in degrees or radians
        by looking at its magnitude. If the values are all less than PI it
        assumes the angles are in radians otherwise it assumes degrees
    angs : (3,) :py:class:`numpy.ndarray` or list
        Lattice angles in degrees or radians as a 3-element array or list
    lattice_vectors : (3, 3) :py:class:`numpy.ndarray` or list of list
        The lattice vectors as a 3x3 matrix, array or list of list
    spacegroup: str or int
        The spacegroup in either International Tables (Hermann-Mauguin)
        notation or a Hall symbol or an integer Hall number.
    is_reciprocal : bool, keyword-only optional (default: False)
        Whether the lattice parameters or lattice vectors refers to a
        reciprocal rather than direct lattice. If True, a/b/c/lens should
        be in reciprocal Angstrom, otherwise they should be in Angstrom
    use_primitive : bool, keyword-only optional (default: True)
        Whether the primitive (or conventional) lattice should be used
    search_length : int, keyword-only optional (default: 1)
        An integer to control how-far the vertex-finding algorithm should
        search in τ-index. The default indicates that (1̄1̄1̄), (1̄1̄0), (1̄1̄1),
        (1̄0̄1), ..., (111) are included.
    time_reversal_symmetry : bool, keyword-only optional (default: False)
        Whether to include time-reversal symmetry as an operation to
        determine the irreducible Brillouin zone
    wedge_search : bool, keyword-only optional (default: True)
        If true, return an irreducible first Brillouin zone,
        otherwise just return the first Brillouin zone
    snap_to_symmetry: bool, keyword-only optional (default: True)
        Enforces that provided lattice parameters / basis vectors / atom-basis
        positions conform to the provided symmetry operations, if present.

    Note
    ----
    Note that the required lattice parameters must be specified as:
        - EITHER ``create_bz(a, b, c, alpha, beta, gamma, spacegroup, ...)``
        - OR     ``create_bz(lens, angs, spacegroup, ...)``
        - OR     ``create_bz(lattice_vectors, spacegroup, ...)``

    E.g. you cannot mix specifing `a`, `b`, `c`, and `angs` etc.
    """
    from brille import Lattice, BrillouinZone

    # Take keyword arguments in preference to positional ones
    a, b, c, alpha, beta, gamma, lens, angs = (
        kwargs.pop(pname, None)
        for pname in ["a", "b", "c", "alpha", "beta", "gamma", "lens", "angs"]
    )
    no_lat_kw_s = any([v is None for v in [a, b, c, alpha, beta, gamma]])
    no_lat_kw_v = any([v is None for v in [lens, angs]])
    if no_lat_kw_v and not no_lat_kw_s:
        lens, angs = ([a, b, c], [alpha, beta, gamma])
    lattice_vectors = kwargs.pop("lattice_vectors", None)
    spacegroup = kwargs.pop("spacegroup", None)
    # Parse positional arguments
    spg_id = 0
    if no_lat_kw_s and no_lat_kw_v and lattice_vectors is None:
        if np.shape(args[0]) == ():
            lens, angs = (args[:3], args[3:6])
            spg_id = 6
        elif np.shape(args[0]) == (3,):
            lens, angs = tuple(args[:2])
            spg_id = 2
        elif np.shape(args[0]) == (3, 1) or np.shape(args[0]) == (1, 3):
            lens, angs = tuple(args[:2])
            lens = np.squeeze(np.array(lens))
            angs = np.squeeze(np.array(angs))
            spg_id = 2
        elif np.shape(args[0]) == (3, 3):
            lattice_vectors = args[0]
            spg_id = 1
        else:
            raise ValueError("No lattice parameters or vectors given")
    if spacegroup is None:
        if len(args) > spg_id:
            spacegroup = args[spg_id]
        else:
            raise ValueError("Spacegroup not given")
    if not isinstance(spacegroup, str):
        e1 = ValueError(
            "Invalid spacegroup input. It must be a string (or valid Spacegroup object)"
        )
        e1.__suppress_context__ = True
        raise e1

    keywords = dict(
        spacegroup=spacegroup,
        real_space=not is_reciprocal,
        snap_to_symmetry=snap_to_symmetry,
    )
    lattice = (
        Lattice((lens, angs), **keywords)
        if lattice_vectors is None
        else Lattice(lattice_vectors, **keywords)
    )

    try:
        return BrillouinZone(
            lattice,
            use_primitive=use_primitive,
            search_length=search_length,
            time_reversal_symmetry=time_reversal_symmetry,
            wedge_search=wedge_search,
        )
    except RuntimeError as e0:
        # We set wedge_search=True by default so add a hint here.
        if "Failed to find an irreducible Brillouin zone" in str(e0):
            e1 = RuntimeError(
                str(e0) + " You can try again with wedge_search=False "
                "to calculate with just the first Brillouin zone"
            )
            e1.__suppress_context__ = True
            e1.__traceback__ = e0.__traceback__
            raise e1
        elif (
            "the dot product between Lattice Vectors requires same or starred lattices"
            in str(e0)
        ):
            e1 = ValueError("Error constructing lattice. Invalid parameters?")
            e1.__suppress_context__ = True
            e1.__traceback__ = e0.__traceback__
            raise e1
        else:
            raise e0


def create_grid(
    bz, complex_values=False, complex_vectors=False, mesh=False, nest=False, **kwargs
):
    """
    Constructs an interpolation grid for a given BrillouinZone object

    Brille provides three different grid implementations:
        - BZTrellisQ: A hybrid Cartesian and tetrahedral grid, with
          tetrahedral nodes on the BZ surface and cuboids inside. [Default]
        - BZMeshQ: A fully tetrahedral grid with a flat data structure
        - BZNestQ: A fully tetrahedral grid with a nested tree data
          structure.

    By default, a BZTrellisQ grid will be used.

    Parameters
    ----------
    bz : :py:class: `BrillouinZone`
        A BrillouinZone object (required)
    complex_values : bool, optional (default: False)
        Whether the interpolated scalar quantities are complex
    complex_vectors : bool, optional (default: False)
        Whether the interpolated vector quantities are complex
    mesh: bool, optional (default: False)
        Whether to construct a BZMeshQ instead of a BZTrellisQ grid
    nest: bool, optional (default: False)
        Whether to construct a BZNestQ instead of a BZTrellisQ grid

    Note
    ----
    Note that setting both `mesh` and `nest` to True gives an error.


    Additional keyword parameters will be passed to the relevant
    grid constructors.

    For ``BZTrellisQ``, these are:

    Parameters
    ----------
    node_volume_fraction : float, optional (default: 1e-5)
        The fractional volume of a tetrahedron in the mesh.
        Smaller numbers will result in better interpolation
        accuracy at the cost of greater computation time.
    always_triangulate : bool, optional (default: False)
        If set to True, we calculate a bounding polyhedron
        for each point in the grid, and triangulate this into
        tetrahedrons. If False, we set internal points to be
        cuboid and compute tetrahedrons only for points near
        the surface of the Brillouin Zone.


    For ``BZMeshQ``, these additional parameters are available:

    Parameters
    ----------
    max_size : float, optional (default: -1.0)
        The maximum volume of a tetrahedron in cubic reciprocal
        Angstrom. If set to a negative value, Tetgen will generate
        a tetrahedral mesh without a volume constraint.
    num_levels : int, optional (default: 3)
        The number of layers of triangulation to use.
    max_points : int, optional (default: -1)
        The maximum number of additional mesh points to add to
        improve the mesh quality. Setting this to -1 will allow
        Tetgen to create an unlimited number of additional points.


    For ``BZNestQ``, these additional parameters are available:

    Parameters
    ----------
    max_volume: float
        Maximum volume of a tetrahedron in cubic reciprocal Angstrom.
    number_density: float
        Number density of points in reciprocal space.
    max_branchings: int, optional (default: 5)
        Maximum number of branchings in the tree structure

    Note
    ----
    Note that one of either the **max_volume** or **number_density**
    parameters must be provided to construct a ``BZNestQ``.
    """
    from brille import BrillouinZone, _brille

    if not isinstance(bz, BrillouinZone):
        raise ValueError("The `bz` input parameter is not a BrillouinZone object")
    if nest and mesh:
        raise ValueError(
            "Both nest=True and mesh=True is set. Please use one or the other"
        )

    def constructor(grid_type):
        if complex_values and complex_vectors:
            return getattr(_brille, grid_type + "cc")
        elif complex_vectors:
            return getattr(_brille, grid_type + "dc")
        else:
            return getattr(_brille, grid_type + "dd")

    if nest:
        if any([v in kwargs for v in ["node_volume_fraction", "always_triangulate"]]):
            raise ValueError(
                "Parameters given are consistent with a trellis grid but nest=True"
            )
        if any([v in kwargs for v in ["max_size", "num_levels", "max_points"]]):
            raise ValueError(
                "Parameters given are consistent with a mesh grid but mesh=False"
            )
        if "max_volume" in kwargs:
            return constructor("BZNestQ")(
                bz, float(kwargs["max_volume"]), kwargs.pop("max_branchings", 5)
            )
        elif "number_density" in kwargs:
            return constructor("BZNestQ")(
                bz, int(kwargs["number_density"]), kwargs.pop("max_branchings", 5)
            )
        else:
            raise ValueError("Neither `max_volume` nor `number_density` provided")
    elif mesh:
        if any([v in kwargs for v in ["node_volume_fraction", "always_triangulate"]]):
            raise ValueError(
                "Parameters given are consistent with a trellis grid but mesh=True"
            )
        if any([v in kwargs for v in ["max_volume", "max_branchings"]]):
            raise ValueError(
                "Parameters given are consistent with a nested grid but nest=False"
            )
        return constructor("BZMeshQ")(
            bz,
            float(kwargs.pop("max_size", -1.0)),
            int(kwargs.pop("num_levels", 3)),
            int(kwargs.pop("max_points", -1)),
        )
    else:
        if any([v in kwargs for v in ["max_volume", "max_branchings"]]):
            raise ValueError(
                "Parameters given are consistent with a nested grid but nest=False"
            )
        if any([v in kwargs for v in ["max_size", "num_levels", "max_points"]]):
            raise ValueError(
                "Parameters given are consistent with a mesh grid but mesh=False"
            )
        return constructor("BZTrellisQ")(
            bz,
            float(kwargs.pop("node_volume_fraction", 1.0e-5)),
            bool(kwargs.pop("always_triangulate", False)),
        )
